from __future__ import annotations


def example_func_prose(
    a: str | None, b: str | int | None = None
) -> dict[str, int]:  # pragma: no cover
    """Example function with a paragraph return section.

    Parameters
    ----------
    a
        An example parameter
    b
        Another, with a default

    Returns
    -------
    An example dict
    """
    return {}


def example_func_tuple() -> tuple[int, str]:  # pragma: no cover
    """Example function with return tuple.

    Returns
    -------
    x
        An example int
    y
        An example str
    """
    return (1, "foo")


def example_func_anonymous_tuple() -> tuple[int, str]:  # pragma: no cover
    """Example function with anonymous return tuple.

    Returns
    -------
    :
        An example int
    :
        An example str
    """
    return (1, "foo")



from __future__ import annotations

import inspect
from types import GenericAlias
from typing import TYPE_CHECKING, TypeAliasType, cast, get_args, get_origin

from sphinx_autodoc_typehints import format_annotation

from scanpydoc import elegant_typehints
from scanpydoc._types import _GenericAlias


if TYPE_CHECKING:
    from typing import Any

    from sphinx.config import Config


def typehints_formatter(annotation: object, config: Config) -> str | None:
    """Generate reStructuredText containing links to the types.

    Can be used as ``typehints_formatter`` for :mod:`sphinx_autodoc_typehints`,
    to respect the ``qualname_overrides`` option.

    Arguments
    ---------
    annotation
        A type or class used as type annotation.
    config
        Sphinx config containing ``sphinx-autodoc-typehints``’s options.

    Returns
    -------
    reStructuredText describing the type
    """
    if isinstance(annotation, TypeAliasType):
        return format_annotation(annotation.__value__, config)

    if inspect.isclass(annotation) and annotation.__module__ == "builtins":
        return None

    tilde = "" if config.typehints_fully_qualified else "~"

    if isinstance(annotation, GenericAlias | _GenericAlias):
        args = get_args(annotation)
        annotation = cast("type[Any]", get_origin(annotation))
    else:
        args = None
    annotation_cls = annotation if inspect.isclass(annotation) else type(annotation)
    if annotation_cls.__module__ in {"typing", "types"}:
        return None

    # Only if this is a real class we override sphinx_autodoc_typehints
    if inspect.isclass(annotation):
        full_name = f"{annotation.__module__}.{annotation.__qualname__}"
        override = elegant_typehints.qualname_overrides.get((None, full_name))
        if override is not None:
            if args is None:
                formatted_args = ""
            else:
                formatted_args = ", ".join(
                    format_annotation(arg, config) for arg in args
                )
                formatted_args = rf"\ \[{formatted_args}]"
            role, qualname = override
            if role is None:
                role = (
                    "py:exc"
                    if issubclass(annotation_cls, BaseException)
                    else "py:class"
                )
            return f":{role}:`{tilde}{qualname}`{formatted_args}"

    return None



"""Prettier function parameter documentation.

This extension replaces the default :class:`~sphinx.domains.python.PyTypedField`
with a derivative :class:`DLTypedField`, which renders item items
(e.g. function parameters) as definition lists instead of simple paragraphs.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from sphinx import addnodes
from docutils import nodes
from sphinx.domains.python import (  # type: ignore[attr-defined,unused-ignore]
    PyObject,
    PyTypedField,
)

from . import metadata, _setup_sig


if TYPE_CHECKING:
    from typing import Any
    from collections.abc import Iterable

    from sphinx.application import Sphinx
    from sphinx.environment import BuildEnvironment

    type TextLikeNode = nodes.Text | nodes.TextElement


class DLTypedField(PyTypedField):
    """A reStructuredText field-list renderer that creates definition lists.

    This style is more readable than cramming name, type, and description in one line.

    The class is not intended to be used directly, but by using the extension,
    it will be used instead of the default :class:`~sphinx.domains.python.PyTypedField`.
    """

    #: Override the list type
    list_type: type[nodes.definition_list] = nodes.definition_list  # type: ignore[assignment]

    def make_field(  # type: ignore[override]
        self,
        types: dict[str, list[nodes.Node]],
        domain: str,
        items: Iterable[tuple[str, list[nodes.inline]]],
        env: BuildEnvironment | None = None,
        **kw: Any,  # noqa: ANN401
    ) -> nodes.field:
        """Render a field to a documenttree node representing a definition list item."""

        def make_refs(
            role_name: str, name: str, node: type[TextLikeNode]
        ) -> list[nodes.Node]:
            return self.make_xrefs(role_name, domain, name, node, env=env, **kw)

        def handle_item(
            fieldarg: str, content: list[nodes.inline]
        ) -> nodes.definition_list_item:
            term = nodes.term()
            term += make_refs(self.rolename, fieldarg, addnodes.literal_strong)

            field_type = types.pop(fieldarg, None)
            if field_type is not None:
                # convert `param : SomeClass` into reference
                if len(field_type) == 1 and isinstance(field_type[0], nodes.Text):
                    classifier_content = make_refs(
                        self.typerolename,
                        field_type[0].astext(),
                        addnodes.literal_emphasis,
                    )
                else:
                    classifier_content = field_type
                term += [
                    # https://github.com/sphinx-doc/sphinx/issues/10815
                    nodes.Text(" "),
                    # Sphinx tries to fixup classifiers without rawsource,
                    # but for this expects attributes we don’t have. Thus “×”.
                    nodes.classifier("×", "", *classifier_content),
                ]

            def_content = nodes.paragraph("", "", *content)
            definition = nodes.definition("", def_content)

            return nodes.definition_list_item("", term, definition)

        field_name = nodes.field_name("", self.label)
        assert not self.can_collapse  # noqa: S101
        body_node = self.list_type(classes=["simple"])
        for field_arg, content in items:
            body_node += handle_item(field_arg, content)
        field_body = nodes.field_body("", body_node)
        return nodes.field("", field_name, field_body)


@_setup_sig
def setup(app: Sphinx) -> dict[str, Any]:
    """Replace :class:`~sphinx.domains.python.PyTypedField` with ours."""
    if (
        "sphinx.ext.napoleon" in app.config.extensions
        and "sphinx.ext.napoleon" not in app.extensions
    ):
        msg = f"Please load sphinx.ext.napoleon before {__name__}"
        raise RuntimeError(msg)

    PyObject.doc_field_types = [
        DLTypedField(
            ft.name,
            names=ft.names,
            typenames=ft.typenames,
            label=ft.label,
            rolename=ft.rolename,
            typerolename=ft.typerolename,
            # Definition lists can’t collapse.
            can_collapse=False,
        )
        if isinstance(ft, PyTypedField)
        else ft
        for ft in PyObject.doc_field_types
    ]

    return metadata



"""A series of Sphinx extensions to get maintainable numpydoc style documentation.

This module is also an extension itself which simply sets up all included extensions.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any
from textwrap import indent
from collections.abc import Callable

from ._version import __version__


if TYPE_CHECKING:
    from sphinx.application import Sphinx


metadata = dict(version=__version__, env_version=1, parallel_read_safe=True)

# Can’t seem to be able to do this in numpydoc style:
# https://github.com/sphinx-doc/sphinx/issues/5887
setup_sig_str = """\
Arguments
---------
app
    Sphinx app to set this :term:`sphinx:extension` up for

Returns
-------
:ref:`Metadata <sphinx:ext-metadata>` for this extension.
"""


def _setup_sig[C: Callable[..., Any]](fn: C) -> C:
    fn.__doc__ = f"{fn.__doc__ or ''}\n\n{indent(setup_sig_str, ' ' * 4)}"
    return fn


@_setup_sig
def setup(app: Sphinx) -> dict[str, Any]:
    """Set up all included extensions!"""  # noqa: D400
    app.setup_extension("scanpydoc.definition_list_typed_field")
    app.setup_extension("scanpydoc.elegant_typehints")
    app.setup_extension("scanpydoc.rtd_github_links")
    app.setup_extension("scanpydoc.theme")
    app.setup_extension("scanpydoc.release_notes")
    return metadata



from __future__ import annotations

import re
import inspect
from types import UnionType
from typing import TYPE_CHECKING, Union, get_args, get_origin, get_type_hints
from typing import Tuple as t_Tuple  # noqa: UP035
from logging import getLogger

from sphinx.ext.napoleon import NumpyDocstring  # type: ignore[attr-defined]
from sphinx_autodoc_typehints import format_annotation


if TYPE_CHECKING:
    from typing import Any
    from collections.abc import Sequence

    from sphinx.application import Sphinx
    from sphinx.ext.autodoc import Options


UNION_TYPES = {Union, UnionType}


__all__ = ["_parse_returns_section", "process_docstring", "setup"]

logger = getLogger(__name__)
re_ret = re.compile("^:returns?: ")


def get_tuple_annot(annotation: type | None) -> tuple[type, ...] | None:
    if annotation is None:
        return None
    origin = get_origin(annotation)
    if not origin:
        return None
    if origin in UNION_TYPES:
        for annot in get_args(annotation):
            origin = get_origin(annot)
            if origin in (tuple, t_Tuple):  # noqa: UP006
                annotation = annot
                break
        else:
            return None
    return get_args(annotation)


def process_docstring(  # noqa: PLR0913
    app: Sphinx,
    what: str,
    name: str,  # noqa: ARG001
    obj: Any,  # noqa: ANN401
    options: Options | None,  # noqa: ARG001
    lines: list[str],
) -> None:
    # Handle complex objects
    if isinstance(obj, property):
        obj = obj.fget
    if not callable(obj):
        return
    if what in ("class", "exception"):
        obj = obj.__init__
    obj = inspect.unwrap(obj)
    try:
        hints = get_type_hints(obj)
    except (AttributeError, NameError, TypeError):  # pragma: no cover
        # Introspecting a slot wrapper can raise TypeError
        return
    ret_types = get_tuple_annot(hints.get("return"))
    if ret_types is None:
        return

    idxs_ret_names = _get_idxs_ret_names(lines)
    if len(idxs_ret_names) == len(ret_types):
        for l, rt in zip(idxs_ret_names, ret_types, strict=False):
            typ = format_annotation(rt, app.config)
            if (line := lines[l]).lstrip() in {":returns: :", ":return: :", ":"}:
                transformed = f"{line[:-1]}{typ}"
            else:
                transformed = f"{line} : {typ}"
            lines[l : l + 1] = [transformed]


def _get_idxs_ret_names(lines: Sequence[str]) -> list[int]:
    # Get return section
    i_prefix = None
    l_start = 0
    for l, line in enumerate(lines):
        if i_prefix is None:
            m = re_ret.match(line)
            if m:
                i_prefix = m.span()[1]
                l_start = l
        elif len(line[:i_prefix].strip()) > 0:
            l_end = l - 1
            break
    else:
        l_end = len(lines) - 1
    if i_prefix is None:
        return []

    # Meat
    idxs_ret_names = []
    for l, line in enumerate([l[i_prefix:] for l in lines[l_start : l_end + 1]]):
        if (line == ":" or line.isidentifier()) and (
            lines[l + l_start + 1].startswith("    ")
        ):
            idxs_ret_names.append(l + l_start)
    return idxs_ret_names


def _parse_returns_section(self: NumpyDocstring, section: str) -> list[str]:  # noqa: ARG001
    """Parse return section as prose instead of tuple by default."""
    lines_raw = list(self._dedent(self._consume_to_next_section()))
    lines = self._format_block(":returns: ", lines_raw)
    if lines and lines[-1]:
        lines.append("")
    return lines


def _delete_sphinx_autodoc_typehints_docstring_processor(app: Sphinx) -> None:
    for listener in app.events.listeners["autodoc-process-docstring"].copy():
        handler_name = getattr(listener.handler, "__name__", None)
        # https://github.com/tox-dev/sphinx-autodoc-typehints/blob/a5c091f725da8374347802d54c16c3d38833d41c/src/sphinx_autodoc_typehints/patches.py#L69
        if handler_name == "napoleon_numpy_docstring_return_type_processor":
            app.disconnect(listener.id)


def setup(app: Sphinx) -> None:
    """Patches the Sphinx app and :mod:`sphinx.ext.napoleon` in some ways.

    1. Replaces the return section parser of napoleon’s NumpyDocstring
       with one that just adds a prose section.
    2. Removes sphinx-autodoc-typehints’s docstring processor that expects
       NumpyDocstring’s old behavior.
    2. Adds our own docstring processor that adds tuple return types
       If the docstring contains a definition list of appropriate length.
    """
    NumpyDocstring._parse_returns_section = _parse_returns_section  # type: ignore[method-assign,assignment]  # noqa: SLF001
    _delete_sphinx_autodoc_typehints_docstring_processor(app)
    app.connect("autodoc-process-docstring", process_docstring, 1000)



"""A release notes directive.

Given a list of version files matching :attr:`FULL_VERSION_RE`,
render them using the following (where ``.`` is the directory they are in):

.. code:: restructuredtext

    .. release-notes:: .

With e.g. the files :file:`1.2.0.md`, :file:`1.2.1.md`, and :file:`1.3.0.rst`,
this will render like the following:

.. code:: restructuredtext

   _v1.3:

   Version 1.3
   ===========

   .. include:: 1.3.0.rst


   _v1.2:

   Version 1.2
   ===========

   .. include:: 1.2.1.md
   .. include:: 1.2.0.md
"""

from __future__ import annotations

import re
import itertools
from typing import TYPE_CHECKING
from pathlib import Path
from dataclasses import dataclass

from docutils import nodes
from packaging.version import Version
from sphinx.util.parsing import nested_parse_to_nodes
from sphinx.util.docutils import SphinxDirective

from . import metadata, _setup_sig


if TYPE_CHECKING:
    from typing import Any, ClassVar
    from collections.abc import Iterable, Sequence

    from sphinx.application import Sphinx
    from myst_parser.mdit_to_docutils.base import DocutilsRenderer


FULL_VERSION_RE = re.compile(r"^(\d+)\.(\d+)\.(\d+)(?:.*)?$")
"""Regex matching a full version number including patch part, maybe with more after."""


@dataclass
class _Backend:
    dir: Path
    instance: SphinxDirective

    def run(self) -> Sequence[nodes.Node]:
        versions = sorted(
            (
                (Version(f.stem), f)
                for f in self.dir.iterdir()
                if FULL_VERSION_RE.match(f.stem)
            ),
            reverse=True,  # descending
        )
        version_groups = itertools.groupby(
            versions, key=lambda vf: (vf[0].major, vf[0].minor)
        )
        return [
            node
            for (major, minor), versions in version_groups
            for node in self.render_version_group(major, minor, versions)
        ]

    def render_version_group(
        self,
        major: int,
        minor: int,
        versions: Iterable[tuple[Version, Path]] = (),
    ) -> tuple[nodes.target, nodes.section]:
        target = nodes.target(
            ids=[f"v{major}-{minor}"],
            names=[f"v{major}.{minor}"],
        )
        section = nodes.section(
            "",
            nodes.title("", f"Version {major}.{minor}"),
            ids=[],
            names=[f"version {major}.{minor}"],
        )

        self.instance.state.document.note_implicit_target(section)
        self.instance.state.document.note_explicit_target(target)

        for _, p in versions:
            section += self.render_include(p)
        return target, section

    def render_include(self, path: Path) -> Sequence[nodes.Node]:
        return nested_parse_to_nodes(
            self.instance.state,
            path.read_text(),
            source=str(path),
            offset=self.instance.content_offset,
        )


# TODO(flying-sheep): Remove once MyST-Parser bug is fixed
# https://github.com/executablebooks/MyST-Parser/issues/967
class _BackendMyst(_Backend):
    def run(self) -> Sequence[nodes.Node]:
        super().run()
        return []

    def render_version_group(
        self, major: int, minor: int, versions: Iterable[tuple[Version, Path]] = ()
    ) -> tuple[nodes.target, nodes.section]:
        target, section = super().render_version_group(major, minor)
        # append target and section to parent
        self._myst_renderer.current_node.append(target)
        self._myst_renderer.update_section_level_state(section, 2)
        # append children to section
        for _, p in versions:
            with self._myst_renderer.current_node_context(section):
                self.render_include(p)
        return target, section  # ignored, just to not change the types

    def render_include(self, path: Path) -> Sequence[nodes.Node]:
        from myst_parser.mocking import MockIncludeDirective
        from docutils.parsers.rst.directives.misc import Include

        srcfile, lineno = self.instance.get_source_info()
        parent_dir = Path(srcfile).parent

        d = MockIncludeDirective(
            renderer=self._myst_renderer,
            name=type(self).__name__,
            klass=Include,
            arguments=[str(path.relative_to(parent_dir))],
            options={},
            body=[],
            lineno=lineno,
        )
        return d.run()

    @property
    def _myst_renderer(self) -> DocutilsRenderer:
        rv: DocutilsRenderer = self.instance.state._renderer  # type: ignore[attr-defined]  # noqa: SLF001
        return rv


class ReleaseNotes(SphinxDirective):
    """Directive rendering release notes, grouping them by minor versions."""

    required_arguments: ClassVar = 1

    def run(self) -> Sequence[nodes.Node]:
        """Read the release notes and render them."""
        dir_ = Path(self.arguments[0])
        # resolve relative dir
        if not dir_.is_absolute():
            src_file = Path(self.get_source_info()[0])
            if not src_file.is_file():
                msg = f"Cannot find relative path to: {src_file}"
                raise self.error(msg)
            dir_ = src_file.parent / self.arguments[0]
        if not dir_.is_dir():
            msg = f"Not a directory: {dir_}"
            raise self.error(msg)

        cls = _BackendMyst if hasattr(self.state, "_renderer") else _Backend
        return cls(dir_, self).run()


@_setup_sig
def setup(app: Sphinx) -> dict[str, Any]:
    """Add the ``release-notes`` directive."""
    app.add_directive("release-notes", ReleaseNotes)
    return metadata



"""Test rtd_github_links subextension."""

from __future__ import annotations

import re
import sys
import textwrap
from typing import TYPE_CHECKING
from pathlib import Path, PurePosixPath
from importlib import import_module
from collections.abc import Mapping

import pytest
from sphinx.config import Config

from scanpydoc.rtd_github_links import (
    _testdata,
    github_url,
    _infer_vars,
    _get_linenos,
    _module_path,
    _get_obj_module,
)
from scanpydoc.rtd_github_links._linkcode import CInfo, PyInfo, linkcode_resolve


if TYPE_CHECKING:
    from types import ModuleType
    from typing import Literal
    from collections.abc import Callable

    from _pytest.monkeypatch import MonkeyPatch

    from scanpydoc.testing import MakeApp
    from scanpydoc.rtd_github_links._linkcode import Domain, DomainInfo


HERE = Path(__file__).parent


@pytest.fixture
def config() -> Config:
    config = Config()
    config.add(
        "rtd_links_prefix", PurePosixPath("."), rebuild="", types=[PurePosixPath]
    )
    return config


@pytest.fixture
def _env(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setattr("scanpydoc.rtd_github_links.github_base_url", "x")


@pytest.fixture(params=[".", "src"])
def prefix(
    monkeypatch: MonkeyPatch, _env: None, request: pytest.FixtureRequest
) -> PurePosixPath:
    pfx = PurePosixPath(request.param)
    monkeypatch.setattr("scanpydoc.rtd_github_links.rtd_links_prefix", pfx)
    return "x" / pfx / "scanpydoc"


@pytest.mark.parametrize(
    ("opt_name", "opt"),
    [
        pytest.param(
            "html_context",
            dict(
                github_user="scverse",
                github_repo="scanpydoc",
                github_version="test-branch",
            ),
            id="html_context",
        ),
        pytest.param(
            "html_theme_options",
            dict(
                repository_url="https://github.com/scverse/scanpydoc",
                repository_branch="test-branch",
            ),
            id="html_theme_options",
        ),
    ],
)
def test_infer_vars(config: Config, opt_name: str, opt: dict[str, str]) -> None:
    config.add(opt_name, {}, rebuild="", types=[dict])
    config[opt_name] = opt
    gbu, rlp = _infer_vars(config)
    assert gbu == "https://github.com/scverse/scanpydoc/tree/test-branch"
    assert rlp


PAT_VAR = r"needs html_context.*'github_user'.*or html_theme_options.*'repository_url'"


@pytest.mark.parametrize(
    ("setup", "pattern"),
    [
        pytest.param(
            [], r"needs “html_context” or “html_theme_options” to be defined", id="none"
        ),
        pytest.param(["html_context"], PAT_VAR, id="html_context"),
        pytest.param(["html_theme_options"], PAT_VAR, id="html_theme_options"),
        pytest.param(["html_context", "html_theme_options"], PAT_VAR, id="both"),
    ],
)
def test_infer_vars_error(config: Config, setup: list[str], pattern: str) -> None:
    for opt_name in setup:
        config.add(opt_name, {}, rebuild="", types=[dict])
    with pytest.raises(ValueError, match=pattern):
        _infer_vars(config)


def test_app(monkeypatch: MonkeyPatch, make_app_setup: MakeApp) -> None:
    filters: dict[str, Callable[..., object]] = {}
    monkeypatch.setattr("scanpydoc.rtd_github_links._init_vars", lambda *_: None)
    monkeypatch.setattr("scanpydoc.rtd_github_links.DEFAULT_FILTERS", filters)
    app = make_app_setup(
        extensions=["sphinx.ext.linkcode", "scanpydoc.rtd_github_links"],
        html_context=dict(
            github_user="scverse", github_repo="scanpydoc", github_version="test-branch"
        ),
    )
    assert app.config["linkcode_resolve"] is linkcode_resolve
    assert filters == dict(github_url=github_url)


@pytest.mark.parametrize(
    ("module", "name", "obj_path"),
    [
        pytest.param(
            *("rtd_github_links", "github_url", "rtd_github_links/__init__.py"),
            id="basic",
        ),
        pytest.param(
            "elegant_typehints",
            "example_func_prose",
            "elegant_typehints/example.py",
            id="reexport",
        ),
    ],
)
def test_as_function(
    prefix: PurePosixPath, module: str, name: str, obj_path: str
) -> None:
    assert github_url(f"scanpydoc.{module}") == str(prefix / module / "__init__.py")
    obj = getattr(import_module(f"scanpydoc.{module}"), name)
    s, e = _get_linenos(obj)
    assert github_url(f"scanpydoc.{module}.{name}") == f"{prefix}/{obj_path}#L{s}-L{e}"


@pytest.mark.parametrize(
    "cls",
    [
        pytest.param(dict, id="dict"),
        pytest.param(
            Mapping,
            marks=pytest.mark.skipif(
                sys.version_info >= (3, 13),
                reason="In Python 3.13+, Mapping is a regular class",
            ),
        ),
    ],
)
def test_no_line_nos_for_unavailable_source(cls: type) -> None:
    start, end = _get_linenos(cls)
    assert start is end is None


def test_get_github_url_only_annotation(prefix: PurePosixPath) -> None:
    """Doesn’t really work but shouldn’t crash either."""
    url = github_url(f"{_testdata.__name__}.TestCls.test_anno")
    path = prefix.parent / Path(*_testdata.__name__.split("."))
    assert url == str(path.with_suffix(".py"))


def test_get_github_url_error() -> None:
    with pytest.raises(KeyError) as exc_info:
        github_url("test.nonexistant.Thingamajig")
    assert exc_info.value.__notes__[0] == "Qualname: 'test.nonexistant.Thingamajig'"


@pytest.mark.parametrize(
    ("obj_path", "obj", "mod", "path_expected"),
    [
        pytest.param(
            "scanpydoc.indent", textwrap.indent, textwrap, "textwrap.py", id="reexport"
        ),
        pytest.param(
            "scanpydoc.rtd_github_links._testdata.test_func",
            _testdata.test_func,
            _testdata,
            "scanpydoc/rtd_github_links/_testdata.py",
            id="func",
        ),
        pytest.param(
            "scanpydoc.rtd_github_links._testdata.test_func_wrap",
            _testdata.test_func_wrap,
            _testdata,
            "scanpydoc/rtd_github_links/_testdata.py",
            id="wrapper",
        ),
        pytest.param(
            "scanpydoc.rtd_github_links._testdata",
            _testdata,
            _testdata,
            "scanpydoc/rtd_github_links/_testdata.py",
            id="mod",
        ),
        pytest.param(
            "scanpydoc.rtd_github_links._testdata.TestDataCls.test_attr",
            _testdata.TestDataCls.__dataclass_fields__["test_attr"],
            _testdata,
            "scanpydoc/rtd_github_links/_testdata.py",
            id="dataclass_field",
        ),
        pytest.param(
            "scanpydoc.rtd_github_links._testdata.TestCls.test_anno",
            None,
            _testdata,
            "scanpydoc/rtd_github_links/_testdata.py",
            id="anno",
        ),
        pytest.param(
            "scanpydoc.rtd_github_links._testdata.TestGenericBuiltin",
            _testdata.TestGenericBuiltin,
            _testdata,
            "scanpydoc/rtd_github_links/_testdata.py",
            id="generic_builtin",
        ),
        pytest.param(
            "scanpydoc.rtd_github_links._testdata.TestGenericClass",
            _testdata.TestGenericClass,
            _testdata,
            "scanpydoc/rtd_github_links/_testdata.py",
            id="generic_class",
        ),
        pytest.param(
            "scanpydoc.rtd_github_links._testdata.TestGenericBuiltinOld",
            _testdata.TestGenericBuiltinOld,
            _testdata,
            "scanpydoc/rtd_github_links/_testdata.py",
            id="generic_builtin_old",
        ),
        pytest.param(
            "scanpydoc.rtd_github_links._testdata.TestGenericClassOld",
            _testdata.TestGenericClassOld,
            _testdata,
            "scanpydoc/rtd_github_links/_testdata.py",
            id="generic_class_old",
        ),
    ],
)
def test_get_obj_module_path(
    obj_path: str, obj: object, mod: ModuleType, path_expected: PurePosixPath
) -> None:
    obj_rcv, mod_rcv = _get_obj_module(obj_path)
    assert obj_rcv is obj
    assert mod_rcv is mod
    path = _module_path(obj_rcv, mod_rcv)
    assert path == PurePosixPath(path_expected)


def test_linkdoc(prefix: PurePosixPath) -> None:
    link = linkcode_resolve(
        "py", PyInfo(module="scanpydoc.rtd_github_links", fullname="setup")
    )
    assert link is not None
    url, hash_ = link.split("#")
    assert url == f"{prefix}/rtd_github_links/__init__.py"
    assert re.fullmatch(r"L\d+-L\d+", hash_)


@pytest.mark.parametrize(
    ("domain", "info"),
    [("py", PyInfo(fullname="foo", module="")), ("c", CInfo(names=[]))],
)
def test_linkcode_skip(domain: Literal[Domain], info: DomainInfo) -> None:
    assert linkcode_resolve(domain, info) is None  # type: ignore[arg-type]



"""Test definition_list_typed_field subextension."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from sphinx import addnodes
from docutils import nodes
from sphinx.testing.restructuredtext import parse


if TYPE_CHECKING:
    from sphinx.application import Sphinx

    from scanpydoc.testing import MakeApp


@pytest.fixture
def app(make_app_setup: MakeApp) -> Sphinx:
    app = make_app_setup()
    app.setup_extension("scanpydoc.definition_list_typed_field")
    return app


# Avoid :class: to not get pending_xref. TODO: fix
params_code_single = """\
.. py:function:: test(a)

   :param a: Only parameter
   :type a: ``str``
"""

params_code = """\
.. py:function:: test(a, b=None)

   :param a: First parameter
   :type a: str
   :param b: Second parameter
   :type b: ``~typing.Optional``\\[``str``]
"""


def test_apps_separate(app: Sphinx, make_app_setup: MakeApp) -> None:
    app_no_setup = make_app_setup()
    assert app is not app_no_setup
    assert "scanpydoc.definition_list_typed_field" in app.extensions
    assert "scanpydoc.definition_list_typed_field" not in app_no_setup.extensions


@pytest.mark.parametrize(
    ("code", "n", "i", "conv_types"),
    [
        pytest.param(params_code_single, 1, 0, [nodes.literal], id="1-simple"),
        pytest.param(params_code, 2, 0, [addnodes.pending_xref], id="2-refconv"),
        pytest.param(
            params_code,
            2,
            1,
            [nodes.literal, nodes.Text, nodes.literal, nodes.Text],
            id="2-multi",
        ),
    ],
)
def test_convert_params(
    app: Sphinx, code: str, n: int, i: int, conv_types: list[type[nodes.Node]]
) -> None:
    # the directive class is PyModuleLevel → PyObject → ObjectDescription
    # ObjectDescription.run uses a DocFieldTransformer to transform members
    # the signature of each Directive(
    #     name, arguments, options, content, lineno,
    #     content_offset, block_text, state, state_machine,

    doc = parse(app, code)
    assert isinstance(desc := doc[1], addnodes.desc)
    assert desc["desctype"] == "function"
    assert isinstance(desc_content := desc[1], addnodes.desc_content)
    assert isinstance(field_list := desc_content[0], nodes.field_list)
    assert isinstance(field := field_list[0], nodes.field)
    assert isinstance(field_name := field[0], nodes.field_name)
    assert isinstance(field_name_text := field_name[0], nodes.Text)
    assert field_name_text.astext() == "Parameters"
    assert isinstance(field_body := field[1], nodes.field_body)
    assert isinstance(dl := field_body[0], nodes.definition_list)

    # each parameter is a dl item that contains a term and a definition
    assert len(dl) == n, dl.children
    assert isinstance(dli := dl[i], nodes.definition_list_item)
    assert isinstance(term := dli[0], nodes.term)
    assert isinstance(dli[1], nodes.definition)

    # the dl term contains the parameter name and type
    assert len(term) == 3, term.children  # noqa: PLR2004
    assert isinstance(term[0], addnodes.literal_strong)
    assert isinstance(term[1], nodes.Text)
    assert term[1].astext() == " "  # spacer
    assert isinstance(cyr := term[2], nodes.classifier)
    assert len(cyr) == len(conv_types), cyr.children
    assert all(
        isinstance(cyr_part, conv_type)
        for cyr_part, conv_type in zip(cyr, conv_types, strict=True)
    )


def test_load_error(make_app_setup: MakeApp) -> None:
    with pytest.raises(RuntimeError, match=r"Please load sphinx\.ext\.napoleon before"):
        make_app_setup(
            extensions=["scanpydoc.definition_list_typed_field", "sphinx.ext.napoleon"]
        )



from __future__ import annotations

from typing import TYPE_CHECKING
from itertools import chain
from collections.abc import MutableMapping


if TYPE_CHECKING:
    from typing import Self
    from collections.abc import Mapping, Iterator


class RoleMapping(MutableMapping[tuple[str | None, str], tuple[str | None, str]]):
    data: dict[tuple[str | None, str], tuple[str | None, str]]

    def __init__(
        self,
        mapping: Mapping[tuple[str | None, str], str | tuple[str | None, str]] = {},
        /,
    ) -> None:
        self.data = dict(mapping)  # type: ignore[arg-type]

    @classmethod
    def from_user(
        cls, mapping: Mapping[str | tuple[str, str], str | tuple[str, str]]
    ) -> Self:
        rm = cls({})
        rm.update_user(mapping)
        return rm

    def update_user(
        self, mapping: Mapping[str | tuple[str, str], str | tuple[str, str]]
    ) -> None:
        for k, v in mapping.items():
            self[k if isinstance(k, tuple) else (None, k)] = (
                v if isinstance(v, tuple) else (None, v)
            )

    def __setitem__(
        self, key: tuple[str | None, str], value: tuple[str | None, str]
    ) -> None:
        self.data[key] = value

    def __getitem__(self, key: tuple[str | None, str]) -> tuple[str | None, str]:
        if key[0] is not None:
            try:
                return self.data[key]
            except KeyError:
                return self.data[None, key[1]]
        for known_role in chain([None], {r for r, _ in self}):
            try:
                return self.data[known_role, key[1]]
            except KeyError:
                pass
        raise KeyError(key)

    def __contains__(self, key: object) -> bool:
        if not isinstance(key, tuple):  # pragma: no cover
            raise TypeError
        try:
            self[key]
        except KeyError:
            return False
        return True

    def __delitem__(self, key: tuple[str | None, str]) -> None:
        del self.data[key]

    def __iter__(self) -> Iterator[tuple[str | None, str]]:
        return self.data.__iter__()

    def __len__(self) -> int:  # pragma: no cover
        return len(self.data)



"""Generate autosummary for imported members.

.. deprecated:: 0.11.0

   Use ``autosummary_imported_members = True`` instead.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from . import _setup_sig


if TYPE_CHECKING:
    from typing import Never

    from sphinx.application import Sphinx


@_setup_sig
def setup(_app: Sphinx) -> Never:  # pragma: no cover
    """Throws an :exc:`ImportError`."""
    msg = (
        "Please use `autosummary_imported_members = True` "
        f"instead of the {__name__} Sphinx extension."
    )
    raise ImportError(msg)



"""Testing utilities."""

from __future__ import annotations

from typing import TYPE_CHECKING, Protocol


if TYPE_CHECKING:
    from typing import Any

    from sphinx.testing.util import SphinxTestApp


class MakeApp(Protocol):
    """Create a SphinxTestApp instance."""

    def __call__(  # noqa: D102
        self,
        builder: str = "html",
        /,
        *,
        exception_on_warning: bool = False,
        **conf: Any,  # noqa: ANN401
    ) -> SphinxTestApp: ...



"""Test release_notes subextension."""

from __future__ import annotations

from types import MappingProxyType
from typing import TYPE_CHECKING
from textwrap import dedent
from functools import partial

import pytest
from sphinx.errors import SphinxWarning
from docutils.utils import new_document
from docutils.languages import get_language
from docutils.parsers.rst import directives


if TYPE_CHECKING:
    from typing import Literal
    from pathlib import Path
    from collections.abc import Mapping

    from sphinx.application import Sphinx
    from sphinx.testing.util import SphinxTestApp

    from scanpydoc.testing import MakeApp

    type Tree = Mapping[str | Path, "Tree | str"]


def mkfiles(root: Path, tree: Tree = MappingProxyType({})) -> None:
    root.mkdir(parents=True, exist_ok=True)
    for path, sub in tree.items():
        if isinstance(sub, str):
            (root / path).write_text(sub)
        else:
            mkfiles(root / path, sub)


@pytest.fixture(params=["rst", "myst"])
def file_format(request: pytest.FixtureRequest) -> Literal["rst", "myst"]:
    return request.param  # type: ignore[no-any-return]


@pytest.fixture
def make_app_relnotes(
    make_app_setup: MakeApp, file_format: Literal["rst", "myst"]
) -> MakeApp:
    return partial(
        make_app_setup,
        "pseudoxml",
        extensions=[
            *(["myst_parser"] if file_format == "myst" else []),
            "scanpydoc.release_notes",
        ],
        exclude_patterns=["[!i]*.md"],
    )


@pytest.fixture
def app(make_app_relnotes: MakeApp) -> SphinxTestApp:
    return make_app_relnotes()


@pytest.fixture
def index_filename(file_format: Literal["rst", "myst"]) -> str:
    return "index.md" if file_format == "myst" else "index.rst"


@pytest.fixture
def index_template(file_format: Literal["rst", "myst"]) -> str:
    return (
        "```{{release-notes}} {}\n```"
        if file_format == "myst"
        else ".. release-notes:: {}"
    )


@pytest.fixture
def files(
    file_format: Literal["rst", "myst"], index_filename: str, index_template: str
) -> Tree:
    files: Tree
    if file_format == "myst":
        files = {
            index_filename: index_template.format("."),
            "1.2.0.md": "(v1.2.0)=\n### 1.2.0",
            "1.2.1.md": "(v1.2.1)=\n### 1.2.1",
            "1.3.0rc1.md": "(v1.3.0rc1)=\n### 1.3.0rc1",
        }
    else:
        files = {
            index_filename: index_template.format("."),
            "1.2.0.rst": ".. _v1.2.0:\n1.2.0\n=====",
            "1.2.1.rst": ".. _v1.2.1:\n1.2.1\n=====",
            "1.3.0rc1.rst": ".. _v1.3.0rc1:\n1.3.0rc1\n========",
        }
    return files


expected = """\
<target refid="v1-3">
<section ids="version-1-3 v1-3" names="version\\ 1.3 v1.3">
    <title>
        Version 1.3
    <target refid="v1-3-0rc1">
    <section ids="rc1 v1-3-0rc1" names="1.3.0rc1 v1.3.0rc1">
        <title>
            1.3.0rc1
<target refid="v1-2">
<section ids="version-1-2 v1-2" names="version\\ 1.2 v1.2">
    <title>
        Version 1.2
    <target refid="v1-2-1">
    <section ids="v1-2-1 id1" names="1.2.1 v1.2.1">
        <title>
            1.2.1
    <target refid="v1-2-0">
    <section ids="v1-2-0 id2" names="1.2.0 v1.2.0">
        <title>
            1.2.0
"""


def test_release_notes(tmp_path: Path, app: Sphinx, files: Tree) -> None:
    mkfiles(tmp_path, files)
    app.build()
    index_out = (tmp_path / "_build/pseudoxml/index.pseudoxml").read_text()
    assert (
        "\n".join(l[4:] for l in dedent(index_out).splitlines()[1:]) == expected.strip()
    )


@pytest.mark.parametrize(
    ("root", "files", "pattern"),
    [
        pytest.param(
            "doesnt-exist.txt", {}, r"Not a directory:.*doesnt-exist.txt", id="nothing"
        ),
        pytest.param(
            "file.txt", {"file.txt": "cont"}, r"Not a directory:.*file.txt", id="file"
        ),
    ],
)
def test_error_wrong_file(
    tmp_path: Path,
    make_app_relnotes: MakeApp,
    index_filename: str,
    index_template: str,
    root: str,
    files: Tree,
    pattern: str,
) -> None:
    app = make_app_relnotes(exception_on_warning=True)
    mkfiles(tmp_path, {index_filename: index_template.format(root), **files})
    with pytest.raises(SphinxWarning, match=pattern):
        app.build()


def test_error_no_src(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
    make_app_relnotes: MakeApp,
    files: Tree,
) -> None:
    app = make_app_relnotes(exception_on_warning=True)
    if "myst_parser" not in app.extensions:
        pytest.skip("rst parser doesn’t need this")
    rn, _ = directives.directive("release-notes", get_language("en"), new_document(""))
    monkeypatch.setattr(rn, "get_source_info", lambda *_, **__: ("<string>", 0))

    mkfiles(tmp_path, files)
    with pytest.raises(SphinxWarning, match=r"Cannot find relative path to: <string>"):
        app.build()



"""This module exists just for rtd_github_links tests."""  # noqa: D404

from __future__ import annotations

from typing import TYPE_CHECKING, Generic, TypeVar
from dataclasses import field, dataclass

from legacy_api_wrap import legacy_api


if TYPE_CHECKING:
    from typing import TypeAlias


class _G[T]:
    pass


_T = TypeVar("_T")


class _G_Old(Generic[_T]):  # noqa: N801, UP046
    pass


# make sure that TestGenericClass keeps its __module__
_G.__module__ = _G_Old.__module__ = "somewhere_else"


type TestGenericBuiltin = list[str]
type TestGenericClass = _G[int]
TestGenericBuiltinOld: TypeAlias = list[str]  # noqa: UP040
TestGenericClassOld: TypeAlias = _G_Old[int]  # noqa: UP040


@dataclass
class TestDataCls:
    test_attr: dict[str, str] = field(default_factory=dict)


class TestCls:
    test_anno: int


def test_func() -> None:  # pragma: no cover
    pass


@legacy_api()
def test_func_wrap() -> None:  # pragma: no cover
    pass
